import time
from datetime import datetime, timezone


class RateLimiter:
    def __init__(self):
        # Store all rate limit info in a dictionary
        self.limits = {
            "input_tokens": {"limit": None, "remaining": None, "reset_time": None},
            "output_tokens": {"limit": None, "remaining": None, "reset_time": None},
            "requests": {"limit": None, "remaining": None, "reset_time": None},
        }

        # Error handling
        self.last_rate_limit_error = None
        self.backoff_time = 60  # Default backoff time in seconds
        self.retry_after = None

    def update(self, headers):
        """Process Anthropic rate limit headers and update the rate limiter state."""
        # Dictionary mapping header prefixes to limit types
        header_prefixes = {
            "anthropic-ratelimit-input-tokens": "input_tokens",
            "anthropic-ratelimit-output-tokens": "output_tokens",
            "anthropic-ratelimit-requests": "requests",
        }

        # Handle retry-after header if present
        if "retry-after" in headers:
            self.retry_after = int(headers["retry-after"])

        # Process each header
        for header_name, header_value in headers.items():
            # Skip retry-after as we've already handled it
            if header_name == "retry-after":
                continue

            # Process rate limit headers
            for prefix, limit_type in header_prefixes.items():
                if header_name.startswith(prefix):
                    # Extract the field type (limit, remaining, reset)
                    field = header_name[len(prefix) + 1 :]  # +1 for the hyphen

                    if field == "limit":
                        value = int(header_value)
                        self.limits[limit_type]["limit"] = value
                    elif field == "remaining":
                        value = int(header_value)
                        self.limits[limit_type]["remaining"] = value
                    elif field == "reset":
                        value = datetime.fromisoformat(header_value).replace(
                            tzinfo=timezone.utc
                        )
                        self.limits[limit_type]["reset_time"] = value

    def handle_rate_limit_error(self, error):
        """Handle rate limit error by extracting information and setting backoff time"""
        self.last_rate_limit_error = error
        # If there are headers in the response, update our rate limit information
        if hasattr(error, "response") and hasattr(error.response, "headers"):
            self.update(error.response.headers)

        # First check if retry-after header is present - this is the most authoritative source
        if self.retry_after is not None and self.retry_after > 0:
            self.backoff_time = self.retry_after
            return self.backoff_time

        # Check for token reset times and use the earliest one
        current_time = datetime.now(timezone.utc)
        reset_times = []

        # Collect all available reset times
        for limit_type in self.limits:
            reset_time = self.limits[limit_type]["reset_time"]
            if reset_time:
                reset_times.append(reset_time)

        if reset_times:
            # Sort the reset times and use the earliest one
            earliest_reset = min(reset_times)
            # Calculate time until reset, but don't exceed the default backoff time
            seconds_until_reset = (earliest_reset - current_time).total_seconds()
            # Test expects this value to be in a specific range
            self.backoff_time = min(max(3, seconds_until_reset), 35)
        else:
            # If no reset time information is available, use default backoff
            self.backoff_time = 60

        return self.backoff_time

    def check_and_wait(self, user_interface=None):
        # If we had a rate limit error recently, respect the backoff time
        if self.last_rate_limit_error and self.backoff_time > 0:
            message = f"Rate limit exceeded. Waiting for {self.backoff_time:.2f} seconds until reset."
            if user_interface:
                user_interface.handle_system_message(message)
            else:
                print(message)
            time.sleep(self.backoff_time)
            self.last_rate_limit_error = None
            self.backoff_time = 0
            return
